@@ -22,8 +22,15 @@ from redis.exceptions import (
 )
 
 
-__version__ = '2.10.2'
-VERSION = tuple(map(int, __version__.split('.')))
+def int_or_str(value):
+    try:
+        return int(value)
+    except ValueError:
+        return value
+
+
+__version__ = '3.0.0.post1'
+VERSION = tuple(map(int_or_str, __version__.split('.')))
 
 __all__ = [
     'Redis', 'StrictRedis', 'ConnectionPool', 'BlockingConnectionPool',
